package org.freecompany.brimstone.main;

import java.io.IOException;
import java.lang.Thread.UncaughtExceptionHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.LinkedList;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Logger;
import org.freecompany.brimstone.core.URNConnection;
import org.freecompany.brimstone.module.ModuleClassLoader;
import org.freecompany.brimstone.module.Module;
import org.freecompany.brimstone.urn.URNStreamHandlerFactory;
import org.freecompany.yacli.CommandLine;
import org.freecompany.yacli.InvalidCommandException;
import org.freecompany.yacli.Option;
import org.freecompany.yacli.Usage;

import static java.lang.System.err;
import static java.lang.System.out;
import static java.util.Arrays.asList;
import static java.util.logging.Level.INFO;
import static java.util.logging.Level.FINE;
import static java.util.logging.Level.WARNING;
import static java.util.logging.Level.SEVERE;
import static java.util.logging.Logger.getLogger;
import static org.freecompany.brimstone.core.URNConnection.addListener;
import static org.freecompany.brimstone.core.Util.installCache;
import static org.freecompany.brimstone.core.Util.isBrimstoneUri;
import static org.freecompany.brimstone.cache.Util.loadToCache;
import static org.freecompany.yacli.Required.OPTIONAL;
import static org.freecompany.yacli.Required.REQUIRED;
import static org.freecompany.yacli.Value.ONE_OR_NONE;
import static org.freecompany.yacli.Value.EXACTLY_ONE;

@Usage( "brimstone [options]")
public class Main implements Runner, UncaughtExceptionHandler {

	private final static Logger logger = getLogger( Main.class.getName());
	private static Main main;

	public static Runner getRunner() {
		return main;
	}

	public static void main( String[] arguments) throws Exception {
		if ( main != null) logger.log( WARNING, "Another instance of the main is active, overwriting.");
		main = new Main();
		main.configure( arguments);
		main.start();
	}
	
	@Option( shortName="p", longName="path", description="Comma separated list of URIs to use as the runtime path.", required=OPTIONAL, value=EXACTLY_ONE)
	public String path;
	
	@Option( shortName="c", longName="class", description="Fully qualified class name to execute.", required=REQUIRED, value=EXACTLY_ONE)
	public String clazz;
	
	@Option( shortName="m", longName="method", description="Static method to be invoked on the specified class, defaults to 'main'.", required=OPTIONAL, value=EXACTLY_ONE)
	public String method = "main";

	protected final AtomicBoolean restart = new AtomicBoolean( true);
	protected ClassLoader parent;
	protected String[] arguments;

	protected void configure( final String[] arguments) throws ClassNotFoundException {
		Class.forName( "org.freecompany.brimstone.module.Module");
		addListener( new DefaultListener( out));
		Thread.setDefaultUncaughtExceptionHandler( this);

		try {
			final CommandLine command = new CommandLine( this);
			command.parse( arguments);
			setArguments( command.remaining());
		} catch ( InvalidCommandException e) {
			err.println( e.getMessage());
			return;
		}

		logger.log( FINE, "Using arguments ''{0}''.", asList( arguments));
		setParent( getClass().getClassLoader());
	}

	public void setPath( final String path) {
		this.path = path;
	}

	public void setClass( final String clazz) {
		this.clazz = clazz;
	}

	public void setMethod( final String method) {
		this.method = method;
	}

	public void setParent( final ClassLoader parent) {
		this.parent = parent;
	}

	public void setArguments( final String[] arguments) {
		this.arguments = arguments;
	}

	public void start() throws ClassNotFoundException, NoSuchMethodException, IllegalAccessException, URISyntaxException, MalformedURLException, IOException {
		while ( restart.getAndSet( false)) {
			try {
				final LinkedList< URL> list = new LinkedList< URL>();
				if ( path != null) {
					for ( String uri : path.split( ",")) {
						URI url = new URI( uri);
						if ( isBrimstoneUri( url)) loadToCache( url);
						list.add( url.toURL());
					}
				}
				final URL[] urls = list.toArray( new URL[ list.size()]);

				logger.log( FINE, "Creating child classloader of ''{0}'' with path ''{1}''.", new Object[] { parent, asList( urls)});
				final ClassLoader loader = new ModuleClassLoader( urls, parent);
				logger.log( FINE, "Loading method ''{0}'' from class ''{1}'' and executing with arguments ''{2}''.", new Object[] { method, clazz, asList( arguments)});
				logger.log( INFO, "Starting application.");
				loader.loadClass( clazz).getMethod( method, String[].class).invoke( null, ( Object) arguments);
			} catch ( InvocationTargetException e) {
				uncaughtException( Thread.currentThread(), e);
			}
		}
	}

	public void restart() {
		if ( !restart.getAndSet( true)) logger.log( INFO, "Initiating restart mode.");
	}

	/**
	 * The default exception handler will simply log a {@link SEVERE} level message
	 * with the exception object.
	 *
	 * @param t the thread which caused the exception.
	 * @param e the exception.
	 */
	public void uncaughtException( final Thread t, final Throwable e) {
		logger.log( SEVERE, "An unhandled exception occurred.", e);
	}
}
