package org.freecompany.brimstone.main;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;

/**
 * Exposes a basic control interface for applications running under the
 * <a href='http://www.freecompany.org/brimstone'>Brimstone</a> main.  The methods exposed allow for basic
 * lifecycle management, allowing for starting, stopping, and restaring
 * the application.
 */
public interface Runner {

	/**
   * Sets the {@link ClassLoader} this main should use as the parent of the {@link ClassLoader} that will be built
   * from the provided path.  This defaults to the classloader of the main implementation itself
   * and if it is overridden applications executed through this main may not to resolve and use
   * this management interface.
   * 
	 * @param parent the {@link ClassLoader} to be used as the parent of the application {@link ClassLoader}.
	 */
	void setParent( ClassLoader parent);

	/**
	 * Comma-separated list of {@link URL} strings that the main will use to look up the {@link Class} and
	 * {@link java.lang.reflect.Method} instances used to load and run the wrapped application.  This set of
	 * locations will also serve as the runtime classpath for the loaded application.
	 *
	 * @param path a list of urls to load from.
	 */
	void setPath( String path);

	/**
	 * Sets the class name that the main will load as the entry point to the application.
	 *
	 * @param clazz class name of the main application.
	 */
	void setClass( String clazz);

	/**
	 * Sets the method name used to start the application.  This method must be a static method on the
	 * {@link Class} loaded by name from the <code>clazz</code> attribute.
	 *
	 * @param method name to invoke when starting the application.
	 */
	void setMethod( String method);

	/**
     * Provides arguments to be passed to the specified static run method.  The default is to not provide any
     * arguments when the main is invoked programmatically.  If the main is being invoked through its main
     * method then the default is to provide all the arguments supplied to the main that do not match known
     * optins in the main itself.
	 *
	 * @param arguments arguments for the main method.
	 */
	void setArguments( String[] arguments);

	/**
	 * Starts the application as described by the three attributes above.
	 *
	 * @throws ClassNotFoundException if the specified class can not be found.
	 * @throws NoSuchMethodException the specified method is not a static method on the given class.
	 * @throws IllegalAccessException the provided class or method cannot be accessed from this application.
	 * @throws URISyntaxException if a
	 * @throws MalformedURLException
	 * @throws IOException 
	 */
	void start() throws ClassNotFoundException, NoSuchMethodException, IllegalAccessException, URISyntaxException, MalformedURLException, IOException ;

	/**
	 * Indicates the main that after the application thread terminates it should be restarted.  Any
   * changes made to the path, class or method will be used when restarting the application.  This
   * method is useful for updating an application either by overwriting the JAR file or by setting
   * up a new path before exiting.
	 */
	void restart();
}
